#include <torch/extension.h>
#include <torch/torch.h>
#include <ATen/ATen.h>
#include <ATen/cuda/CUDAContext.h>
#include <vector>
#include <tuple>

// CUDA kernel to perform coalesced copy
template <typename scalar_t>
__global__ void coalesced_copy_kernel(const scalar_t* __restrict__ src, scalar_t* __restrict__ dst, size_t numel) {
    size_t idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx < numel) {
        dst[idx] = src[idx];
    }
}

// Helper function to launch the custom copy kernel ensuring memory coalescing
void coalesced_copy_tensor(const torch::Tensor &src, torch::Tensor &dst) {
    auto numel = src.numel();
    int threads = 256;
    int blocks = (numel + threads - 1) / threads;
    AT_DISPATCH_FLOATING_TYPES(src.scalar_type(), "coalesced_copy", ([&] {
        coalesced_copy_kernel<scalar_t><<<blocks, threads, 0, at::cuda::getCurrentCUDAStream()>>>(
            src.data_ptr<scalar_t>(),
            dst.data_ptr<scalar_t>(),
            numel
        );
    }));
}

// Main forward function for the optimized LTSMHn kernel with memory coalescing
torch::Tensor forward(
    torch::Tensor x,
    std::vector<torch::Tensor> lstm_weights_ih,
    std::vector<torch::Tensor> lstm_weights_hh,
    std::vector<torch::Tensor> lstm_biases_ih,
    std::vector<torch::Tensor> lstm_biases_hh,
    torch::Tensor h0,
    torch::Tensor c0,
    bool is_training
) {
    // Ensure all tensors are on the proper device and are contiguous
    auto device = x.device();
    x = x.to(device).contiguous();
    h0 = h0.to(device).contiguous();
    c0 = c0.to(device).contiguous();

    // Clone initial states for manipulation
    torch::Tensor out = x;
    torch::Tensor hn = h0.clone();
    torch::Tensor cn = c0.clone();

    const size_t num_layers = lstm_weights_ih.size();

    // Lambda to process each LSTM layer
    auto process_layer = [&](size_t i) {
        // Extract parameters for layer i
        auto weight_ih = lstm_weights_ih[i];
        auto weight_hh = lstm_weights_hh[i];
        auto bias_ih = lstm_biases_ih[i];
        auto bias_hh = lstm_biases_hh[i];

        int64_t input_size = weight_ih.size(1);
        int64_t hidden_size = weight_hh.size(1);

        // Create a one-layer LSTM module
        torch::nn::LSTM lstm_model(
            torch::nn::LSTMOptions(input_size, hidden_size)
                .num_layers(1)
                .batch_first(true)
                .bidirectional(false)
        );
        lstm_model->to(device);

        // Copy the layer parameters (assumed to be stored contiguously)
        lstm_model->named_parameters()["weight_ih_l0"].copy_(weight_ih);
        lstm_model->named_parameters()["weight_hh_l0"].copy_(weight_hh);
        lstm_model->named_parameters()["bias_ih_l0"].copy_(bias_ih);
        lstm_model->named_parameters()["bias_hh_l0"].copy_(bias_hh);

        // Extract the corresponding hidden and cell state slices
        torch::Tensor h_slice = hn.narrow(0, i, 1);
        torch::Tensor c_slice = cn.narrow(0, i, 1);
        std::tuple<torch::Tensor, torch::Tensor> state_tuple = std::make_tuple(h_slice, c_slice);

        lstm_model->train(is_training);

        // Forward pass through the current LSTM layer
        auto output_and_state = lstm_model->forward(out, state_tuple);
        auto output = std::get<0>(output_and_state);
        auto state = std::get<1>(output_and_state);
        auto h_n = std::get<0>(state);
        auto c_n = std::get<1>(state);

        // Ensure outputs are contiguous for coalesced memory access
        h_n = h_n.contiguous();
        c_n = c_n.contiguous();

        // Update the global hidden and cell states using the coalesced copy kernel
        torch::Tensor h_target = hn.narrow(0, i, 1);
        torch::Tensor c_target = cn.narrow(0, i, 1);
        coalesced_copy_tensor(h_n, h_target);
        coalesced_copy_tensor(c_n, c_target);

        // The output of the current layer becomes the input for the next layer
        out = output;
    };

    // Explicitly unroll the first four layers if present
    if (num_layers > 0) process_layer(0);
    if (num_layers > 1) process_layer(1);
    if (num_layers > 2) process_layer(2);
    if (num_layers > 3) process_layer(3);
    for (size_t i = 4; i < num_layers; ++i) {
        process_layer(i);
    }

    return hn;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Optimized LSTM forward (CUDA) with memory coalescing");
}
